@@ -34,7 +34,6 @@
 #include <libspi/spi-private.h>
 #include "accessible.h"
 #include "application.h"
-
 #include <bonobo-activation/bonobo-activation-register.h>
 
 #undef SPI_BRIDGE_DEBUG
@@ -77,7 +76,6 @@ static void     spi_atk_bridge_toplevel_removed        (AtkObject             *o
 
 static void     spi_atk_bridge_exit_func               (void);
 static void     spi_atk_register_event_listeners       (void);
-static void     spi_atk_deregister_event_listeners     (void);
 static void     spi_atk_bridge_focus_tracker           (AtkObject             *object);
 static void     spi_atk_bridge_register_application    (Accessibility_Registry registry);
 static gboolean spi_atk_bridge_property_event_listener (GSignalInvocationHint *signal_hint,
@@ -161,6 +159,8 @@ atk_bridge_init (gint *argc, gchar **argv[])
       g_error ("Could not initialize Bonobo");
     }
 
+  /* Create the accessible application server object */
+  this_app = spi_application_new (atk_get_root ());
   /*
    * We only want to enable the bridge for top level
    * applications, we detect bonobo components by seeing
@@ -182,8 +182,9 @@ atk_bridge_init (gint *argc, gchar **argv[])
   else
     {
       spi_atk_bridge_do_registration ();
-      spi_atk_bridge_init_event_type_consts ();
     }
+  spi_atk_register_event_listeners ();
+  spi_atk_bridge_init_event_type_consts ();
  
   return 0;
 }
@@ -204,8 +205,8 @@ spi_atk_bridge_do_registration (void)
   bonobo_activate ();
 
   /* Create the accessible application server object */
-
-  this_app = spi_application_new (atk_get_root ());
+  if (this_app == NULL)
+    this_app = spi_application_new (atk_get_root ());
 
   DBG (1, g_message ("About to register application\n"));
 
@@ -225,7 +226,6 @@ spi_atk_bridge_toplevel_added (AtkObject *object,
   if (toplevels == 0)
     {
       spi_atk_bridge_do_registration ();
-      spi_atk_bridge_init_event_type_consts ();
     }
   toplevels++;
 }
@@ -241,7 +241,6 @@ spi_atk_bridge_toplevel_removed (AtkObject *object,
   if (toplevels == 0)
     {
       deregister_application (app);
-      spi_atk_deregister_event_listeners ();
       reinit_register_vars ();
     }
   if (toplevels < 0)
@@ -257,7 +256,6 @@ spi_atk_bridge_register_application (Accessibility_Registry registry)
   Accessibility_Registry_registerApplication (spi_atk_bridge_get_registry (),
                                               BONOBO_OBJREF (this_app),
                                               &ev);
-  spi_atk_register_event_listeners ();
 }
 
 static Accessibility_Registry
@@ -411,32 +409,6 @@ spi_atk_register_event_listeners (void)
   g_object_unref (ao);
 }
 
-static void
-spi_atk_deregister_event_listeners (void)
-{
-  gint i;
-  guint id;
-
-  if (!atk_listeners_registered)
-    return;
-
-  atk_listeners_registered = FALSE;
-
-  for (i = 0; i < listener_ids->len; i++)
-    {
-      id = g_array_index (listener_ids, guint, i);
-      atk_remove_global_event_listener (id);   
-    }
-  g_array_free (listener_ids, TRUE);
-  listener_ids = NULL;
-
-  atk_remove_focus_tracker (atk_bridge_focus_tracker_id);
-  atk_bridge_focus_tracker_id = 0;
-
-  atk_remove_key_event_listener (atk_bridge_key_event_listener_id);
-  atk_bridge_key_event_listener_id = 0;
-}
-
 static void
 deregister_application (BonoboObject *app)
 {
