@@ -55,12 +55,14 @@
 #include "../libspi/spi-private.h"
 #include "deviceeventcontroller.h"
 
+KeySym ucs2keysym (long ucs);
+
 #define CHECK_RELEASE_DELAY 20
 #define BIT(c, x)       (c[x/8]&(1<<(x%8)))
 static guint check_release_handler = 0;
 static Accessibility_DeviceEvent pressed_event;
 static SpiDEController *saved_controller; 
-static void wait_for_release_event (GdkEvent *event, SpiDEController *controller);
+static void wait_for_release_event (XEvent *event, SpiDEController *controller);
 
 /* Our parent Gtk object type */
 #define PARENT_TYPE BONOBO_TYPE_OBJECT
@@ -568,17 +570,25 @@ spi_dec_init_mouse_listener (SpiRegistry *registry)
 /**
  * Eventually we can use this to make the marshalling of mask types
  * more sane, but for now we just use this to detect 
- * the use of 'virtual' masks such as Mumlock and convert them to
+ * the use of 'virtual' masks such as numlock and convert them to
  * system-specific mask values (i.e. ModMask).
  * 
  **/
 static Accessibility_ControllerEventMask
 spi_dec_translate_mask (Accessibility_ControllerEventMask mask)
 {
-  if (mask == SPI_KEYMASK_NUMLOCK) {
-    mask = _numlock_physical_mask;
-  }
-  return mask;
+  Accessibility_ControllerEventMask tmp_mask;
+  gboolean has_numlock;
+
+  has_numlock = (mask & SPI_KEYMASK_NUMLOCK);
+  tmp_mask = mask;
+  if (has_numlock)
+    {
+      tmp_mask = mask ^ SPI_KEYMASK_NUMLOCK;
+      tmp_mask |= _numlock_physical_mask;
+    }
+ 
+  return tmp_mask;
 }
 
 static DEControllerKeyListener *
@@ -1012,15 +1022,38 @@ global_filter_fn (GdkXEvent *gdk_xevent, GdkEvent *event, gpointer data)
     {
       if (controller->xevie_display == NULL)
         {
-          gboolean is_consumed =
+          gboolean is_consumed;
+
+          is_consumed =
             spi_device_event_controller_forward_key_event (controller, xevent);
 
           if (is_consumed)
-            XAllowEvents (spi_get_display (), AsyncKeyboard, CurrentTime);
+            {
+              int n_events;
+              int i;
+              XEvent next_event;
+              n_events = XPending (display);
+
+#ifdef SPI_KEYEVENT_DEBUG
+              g_print ("Number of events pending: %d\n", n_events);
+#endif
+              for (i = 0; i < n_events; i++)
+                {
+                  XNextEvent (display, &next_event);
+		  if (next_event.type != KeyPress &&
+		      next_event.type != KeyRelease)
+			g_warning ("Unexpected event type %d in queue", next_event.type);
+                 }
+
+              XAllowEvents (display, AsyncKeyboard, CurrentTime);
+              if (n_events)
+                XUngrabKeyboard (display, CurrentTime);
+            }
           else
             {
-              wait_for_release_event (event, controller);
-              XAllowEvents (spi_get_display (), ReplayKeyboard, CurrentTime);
+              if (xevent->type == KeyPress)
+                wait_for_release_event (xevent, controller);
+              XAllowEvents (display, ReplayKeyboard, CurrentTime);
             }
         }
 
@@ -1034,7 +1067,7 @@ global_filter_fn (GdkXEvent *gdk_xevent, GdkEvent *event, gpointer data)
     {
       XkbAnyEvent * xkb_ev = (XkbAnyEvent *) xevent;
       /* ugly but probably necessary...*/
-      XSynchronize (spi_get_display (), TRUE);
+      XSynchronize (display, TRUE);
 
       if (xkb_ev->xkb_type == XkbStateNotify)
         {
@@ -1096,7 +1129,7 @@ global_filter_fn (GdkXEvent *gdk_xevent, GdkEvent *event, gpointer data)
 	}
         else
 	       DBG (2, g_warning ("XKB event %d\n", xkb_ev->xkb_type));
-      XSynchronize (spi_get_display (), FALSE);
+      XSynchronize (display, FALSE);
     }
   
   return GDK_FILTER_CONTINUE;
@@ -1483,6 +1516,7 @@ spi_keystroke_from_x_key_event (XKeyEvent *x_key_event)
      (x_key_event->state & Mod1Mask)?"Alt-":"",
      ((x_key_event->state & ShiftMask)^(x_key_event->state & LockMask))?
      g_ascii_toupper (keysym) : g_ascii_tolower (keysym));
+  fprintf (stderr, "serial: %x Time: %x\n", x_key_event->serial, x_key_event->time);
 #endif /* SPI_DEBUG */
   return key_event;	
 }
@@ -1493,6 +1527,7 @@ spi_controller_update_key_grabs (SpiDEController           *controller,
 {
   GList *l, *next;
   gboolean   update_failed = FALSE;
+  KeyCode keycode;
   
   g_return_val_if_fail (controller != NULL, FALSE);
 
@@ -1505,6 +1540,8 @@ spi_controller_update_key_grabs (SpiDEController           *controller,
    *
    * ControlMask grabs are broken, must be in use already
    */
+  if (recv)
+    keycode = keycode_for_keysym (recv->id, NULL);
   for (l = controller->keygrabs_list; l; l = next)
     {
       gboolean do_remove;
@@ -1515,7 +1552,7 @@ spi_controller_update_key_grabs (SpiDEController           *controller,
 
       re_issue_grab = recv &&
 	      (recv->modifiers & grab_mask->mod_mask) &&
-	      (grab_mask->key_val == keycode_for_keysym (recv->id, NULL));
+	      (grab_mask->key_val == keycode);
 
 #ifdef SPI_DEBUG
       fprintf (stderr, "mask=%lx %lx (%c%c) %s\n",
@@ -1956,45 +1993,7 @@ dec_unlock_modifiers (SpiDEController *controller, unsigned modifiers)
 static KeySym
 dec_keysym_for_unichar (SpiDEController *controller, gunichar unichar)
 {
-	/* TODO: table lookups within a range, for various code pages! */
-	KeySym keysym = NoSymbol;
-
-	if (unichar >= 0x20 && unichar < 0x7f) { /* Basic Latin/ASCII */
-		keysym = (KeySym) unichar;
-	}
-	else if (unichar >= 0xa0 && unichar <= 0xff) { /* Latin 1 extensions */
-		keysym = (KeySym) unichar;
-	}
-	else if (unichar >= 0x100 && unichar <= 0x233) { /* unfortunately the mapping gets nasty for Latin-2 and 3... help! */
-		keysym = NoSymbol;
-	}
-	else if (unichar >= 0x7c1 && unichar <= 0x3a1) { /* let's try Greek anyway... */
-		keysym = (KeySym) (0x391 + (unichar - 0x7c1));
-	}
-	else if (unichar >= 0x3a3 && unichar <= 0x3a9) { /* let's try Greek anyway... */
-		keysym = (KeySym) (0x7d2 + (unichar - 0x3a3));
-	}
-	else if (unichar >= 0x3b1 && unichar <= 0x3c1) { /* let's try Greek anyway... */
-		keysym = (KeySym) (0x7e1 + (unichar - 0x3b1));
-	}
-	else if (unichar == 0x3c2) {
-		keysym = (KeySym) 0x7f3; /* XK_Greek_finalsmallsigma; */
-	}
-	else if (unichar >= 0x3c3 && unichar <= 0x3c9) { /* let's try Greek anyway... */
-		keysym = (KeySym) (0x7f2 + (unichar - 0x3c2));
-	}	
-	else if (unichar >= 0x5d0 && unichar <= 0x5ea) { /* Hebrew basics */
-		/* probably broken :-) */
-		keysym = (KeySym) (0xce0 + (unichar - 0x5d0));
-	}	
-	else if (unichar >= 0x30a1 && unichar <= 0x30ab) { /* partial katakana support */
-		/* TODO: complete! */
-		keysym = (KeySym) (0x4b1 + (unichar - 0x30a1)/2);
-	}
-	else if (unichar >= 0x20a0 && unichar <= 0x20ac) { /* currency */
-		keysym = (KeySym) unichar; /* how convenient ;-) */
-	}
-	return keysym;
+	return ucs2keysym ((long) unichar);
 }
 
 static gboolean
@@ -2423,7 +2422,7 @@ check_release (gpointer data)
   return (released == 0);
 }
 
-static void wait_for_release_event (GdkEvent        *event,
+static void wait_for_release_event (XEvent          *event,
                                     SpiDEController *controller)
 {
   pressed_event = spi_keystroke_from_x_key_event ((XKeyEvent *) event);
